//
//  DataManager.m
//  DemoRoom
//
//  Created by wang  chao on 12-1-5.
//  Copyright 2012年 bupt. All rights reserved.
//

#import "DataManager.h"
#import "LocalDriver.h"

static NSString* _head = @"_datamanager_";

#define MIN_DATA_FETCH_TIME 5.0

static DataManager* _manager;

@interface DataManager (Private)

- (NSString*)_keyPath:(DataMapper*)mapper params:(NSDictionary*)params;
- (NSString*)_keyHead:(NSString*)key;
- (StoreDriver*)_storeDriverForType:(DataCacheType)type;
- (void)_upToTime;

@end

@implementation DataManager (Private)

- (NSString*)_keyHead:(NSString *)key{
    return [_head stringByAppendingFormat:@"_%@",key];
}

- (NSString*)_keyPath:(DataMapper *)mapper params:(NSDictionary *)params{
    NSString* key = mapper.key;
    return [NSString stringWithFormat:@"%@_%@",[self _keyHead:key],[mapper keyPathForParams:[DataMapper normalParams:params]]];
}

- (StoreDriver*)_storeDriverForType:(DataCacheType)type{
    static StoreDriver* memDriver = nil;
    static StoreDriver* diskDriver = nil;
    if (!memDriver) {
        memDriver = [[LocalDriver alloc] initWithType:MEM];
    }
    
    if (!diskDriver) {
        diskDriver = [[LocalDriver alloc] initWithType:DISK];
    }
    
    if (type == CACHE_MEM) {
        return memDriver;
    }
    
    if (type == CACHE_DISK) {
        return diskDriver;
    }
    return nil;
}

- (void)_upToTime{
    for (NSString* infoKey in _autoList) {
        DataInfo* info = [_autoList objectForKey:infoKey];
        if ([info expired]) {
            [self getValue:info.key params:info.params handler:nil useCache:NO];
        }
    }
}

@end

@implementation DataManager

@synthesize mappers = _mappers;
@synthesize timer = _timer;
@synthesize autoList = _autoList;

#pragma mark - construct methods

- (id)init
{
    self = [super init];
    if (self) {
        self.mappers = [NSMutableDictionary dictionaryWithCapacity:100];
        self.timer = [NSTimer scheduledTimerWithTimeInterval:MIN_DATA_FETCH_TIME/2.0 target:self selector:@selector(_upToTime) userInfo:nil repeats:YES];
        self.autoList = [NSMutableDictionary dictionaryWithCapacity:10];
    }
    
    return self;
}

+ (id)manager{
    if (!_manager) {
        _manager = [[self alloc] init];
    }
    return _manager;
}

- (NSString*)keyPath:(NSString*)key params:(NSDictionary*)params{
    DataMapper* mapper = [self.mappers objectForKey:key];
    if (!mapper) {
        return nil;
    }
    return [self _keyPath:mapper params:params];
}

- (void)addObserver:(id)observer selector:(SEL)selector forKey:(NSString *)key params:(NSDictionary *)params{
    DataMapper* mapper = [self.mappers objectForKey:key];
    if (!mapper) {
        return;
    }
    NSString* obKey = [self _keyPath:mapper params:params];
    [[NSNotificationCenter defaultCenter] addObserver:observer selector:selector name:obKey object:self];
}

- (void)addObserver:(id)observer selector:(SEL)selector forKey:(NSString *)key{
    DataMapper* mapper = [self.mappers objectForKey:key];
    if (!mapper) {
        return;
    }
    NSString* obKey = [self _keyHead:key];
    [[NSNotificationCenter defaultCenter] addObserver:observer selector:selector name:obKey object:self];
}

- (void)removeObserver:(id)observer forKey:(NSString *)key params:(NSDictionary *)params{
    DataMapper* mapper = [self.mappers objectForKey:key];
    if (!mapper) {
        return;
    }
    NSString* obKey = [self _keyPath:mapper params:params];
    [[NSNotificationCenter defaultCenter] removeObserver:observer name:obKey object:self];
}

- (void)removeObserver:(id)observer forKey:(NSString *)key{
    DataMapper* mapper = [self.mappers objectForKey:key];
    if (!mapper) {
        return;
    }
    NSString* obKey = [self _keyHead:key];
    [[NSNotificationCenter defaultCenter] removeObserver:observer name:obKey object:self];
}

- (void)removeObserver:(id)observer{
    [[NSNotificationCenter defaultCenter] removeObserver:observer];
}


#pragma mark - DataMapperDelegate

- (void)mapperStateChanged:(DataMapper *)mapper params:(NSDictionary *)params{
    DataOption* option = [mapper option:params];
    if (!option.needsCache) {
        StoreDriver* driver = [self _storeDriverForType:option.cacheType];
        [driver deleteData:mapper params:params];
    }
    [self refershAutoList];
}

#pragma mark -

- (void)refershAutoList{
    [self.autoList removeAllObjects];
    for (NSString* key in _mappers) {
        DataMapper* mapper = [_mappers objectForKey:key];
        for (NSString* pkey in mapper.autoList) {
            NSDictionary* params = [mapper.autoList objectForKey:pkey];
            DataOption* option = [mapper option:params];
                
            NSString* keyPath = [self _keyPath:mapper params:params];
            StoreDriver* driver = [self _storeDriverForType:option.cacheType];
            DataInfo* info = [driver getData:mapper params:params];
            
            if (info) {
                info.expireAfter = info.expireAfter > MIN_DATA_FETCH_TIME?info.expireAfter:MIN_DATA_FETCH_TIME;
            }else{
                info = [[[DataInfo alloc] init] autorelease];
                info.updateTime = 0;
                info.expireAfter = MIN_DATA_FETCH_TIME;
            }
            info.key = mapper.key;
            info.params = params;
            [_autoList setObject:info forKey:keyPath];
        }
    }
}

- (void)addMapper:(DataMapper *)mapper{
    if (!mapper.key) {
        return;
    }
    [_mappers setObject:mapper forKey:mapper.key];
    mapper.delegate = self;
    [self refershAutoList];
}

- (void)getValue:(NSString *)key params:(NSDictionary *)params handler:(DataHandler)handler useCache:(BOOL)useCache{
    DataMapper* mapper = [_mappers objectForKey:key];    
    if (!mapper) {
        NSError* error = [NSError errorWithDomain:@"DataManager" code:DATA_ERROR_NOMAPPER userInfo:nil];
        if (handler) {
            DataInfo* info = [[[DataInfo alloc] initWithKey:key params:params] autorelease];
            handler(info,error);
        }
        return;
    }
    
    DataOption* option =  [mapper option:params];    
    StoreDriver* driver = [self _storeDriverForType:option.cacheType];
    
    if (useCache) {
        DataInfo* info = [self getCache:key params:params];
        if (info) {
            handler(info,nil);
            return;
        }
    }

    
    [mapper getValue:params handler:^(DataInfo* info,NSError* error){
        info.key = key;
        info.params = params;
        if (!error) {
            if (option.needsCache) {
                [driver store:info mapper:mapper params:params];
            }
            NSString* keyPath = [self _keyPath:mapper params:params];
            DataInfo* autoInfo = [_autoList objectForKey:keyPath];
            if (autoInfo) {
                autoInfo.updateTime = [[NSDate date] timeIntervalSince1970];
                autoInfo.expireAfter = (info.expireAfter > MIN_DATA_FETCH_TIME)?info.expireAfter:MIN_DATA_FETCH_TIME;
            }
        }
        
        info.data = mapper.decoder(info.data);
        
        //handler notificate
        NSString* obkey = [self _keyPath:mapper params:params];
        NSMutableDictionary* userInfo = [NSMutableDictionary dictionaryWithCapacity:2];
        [userInfo setObject:key forKey:@"key"];
        [userInfo setObject:params forKey:@"params"];
        [userInfo setObject:info forKey:@"info"];
        [userInfo setObject:[NSNumber numberWithBool:NO] forKey:@"keyNotification"];
        [[NSNotificationCenter defaultCenter] postNotificationName:obkey object:self userInfo:userInfo];
        NSString* obkey2 = [self _keyHead:key];
        NSMutableDictionary* userInfo2 = [NSMutableDictionary dictionaryWithDictionary:userInfo];
        [userInfo setObject:[NSNumber numberWithBool:YES] forKey:@"keyNotification"];
        [[NSNotificationCenter defaultCenter] postNotificationName:obkey2 object:self userInfo:userInfo2];

        if (handler) {
            handler(info,error);
        }
    }];
}

- (void)getValue:(NSString*)key params:(NSDictionary*)params handler:(DataHandler)handler{
    [self getValue:key params:params handler:handler useCache:YES];
}

- (DataInfo*)getValue:(NSString *)key params:(NSDictionary *)params useCache:(BOOL)useCache{
    DataInfo* dinfo = [[[DataInfo alloc] initWithKey:key params:params] autorelease];
    NSMutableDictionary* done = [NSMutableDictionary dictionaryWithObject:[NSNumber numberWithBool:NO] forKey:@"done"];
    [self getValue:key params:params handler:^(DataInfo* info,NSError* error){
        [dinfo loadInfo:info];
        if (error) {
            [done setObject:[NSNumber numberWithBool:YES] forKey:@"error"];
        }else{
            [done setObject:[NSNumber numberWithBool:NO] forKey:@"error"];
        }
        [done setObject:[NSNumber numberWithBool:YES] forKey:@"done"];
    } useCache:useCache];
    while (![[done objectForKey:@"done"] boolValue]) {
        [[NSRunLoop currentRunLoop] runUntilDate:[NSDate dateWithTimeIntervalSinceNow:0.1]];
    }
    if (![[done objectForKey:@"error"] boolValue]) {
        return dinfo;
    }
    return nil;
}

- (DataInfo*)getValue:(NSString*)key params:(NSDictionary*)params{
    return [self getValue:key params:params useCache:YES];
}

- (void)updateValue:(NSString *)key params:(NSDictionary *)params value:(id)value handler:(DataHandler)handler{
    DataMapper* mapper = [_mappers objectForKey:key];
    
    if (!mapper) {
        NSError* error = [NSError errorWithDomain:@"DataManager" code:DATA_ERROR_NOMAPPER userInfo:nil];
        if (handler) {
            DataInfo* info = [[[DataInfo alloc] initWithKey:key params:params] autorelease];
            handler(info,error);
        }
        return;
    }
    
    DataOption* option =  [mapper option:params];    
    StoreDriver* driver = [self _storeDriverForType:option.cacheType];

    
    [mapper updateValue:params value:value handler:^(DataInfo* info,NSError* error){
        if (!error) {
            if (info.data && option.needsCache) {
                [driver store:info mapper:mapper params:params];
            }else{
                [driver deleteData:mapper params:params];
            }
            NSString* keyPath = [self _keyPath:mapper params:params];
            DataInfo* autoInfo = [_autoList objectForKey:keyPath];
            if (autoInfo) {
                autoInfo.updateTime = [[NSDate date] timeIntervalSince1970];
                autoInfo.expireAfter = (info.expireAfter > MIN_DATA_FETCH_TIME)?info.expireAfter:MIN_DATA_FETCH_TIME;
            }
        }
        info.key = info.key;
        info.params = info.params;
        if (handler) {
            handler(info,error);
        }
    }];
}

- (void)getCacheAndValue:(NSString *)key params:(NSDictionary *)params handler:(DataHandler)handler{
    DataInfo* info = [self getCache:key params:params];
    if (info) {
        if (handler) {
            handler(info,nil);
        }
    }
    [self getValue:key params:params handler:handler useCache:NO];
}

- (DataInfo*)getCache:(NSString *)key params:(NSDictionary *)params{
    DataMapper* mapper = [_mappers objectForKey:key];
    DataInfo* info = nil;
    
    if (!mapper) {
        return nil;
    }
    
    DataOption* option =  [mapper option:params];
    StoreDriver* driver = [self _storeDriverForType:option.cacheType];
    
    if (!option.needsCache) {
        return nil;
    }
    
    info = [driver getData:mapper params:params];
    info.key = key;
    info.params = params;
    if (option.setExpireAfter) {
        info.expireAfter = option.expireAfter;
    }
    if (info) {
        if (![info expired]) {
            info.data = mapper.decoder(info.data);
        }else{
            info = nil;
        }
    }
    info.cache = YES;
    return info;
}

- (void)cleanCache{
    [[self _storeDriverForType:CACHE_MEM] clean];
    [[self _storeDriverForType:CACHE_DISK] clean];
}

- (void)cleanCache:(NSString *)key{
    DataMapper* mapper = [self.mappers objectForKey:key];
    [[self _storeDriverForType:CACHE_MEM] clean:mapper];
    [[self _storeDriverForType:CACHE_DISK] clean:mapper];
}

- (void)cleanCache:(NSString *)key params:(NSDictionary *)params{
    DataMapper* mapper = [self.mappers objectForKey:key];
    [[self _storeDriverForType:CACHE_MEM] deleteData:mapper params:params];
    [[self _storeDriverForType:CACHE_DISK] deleteData:mapper params:params];
}


#pragma mark - manage memory

- (void)dealloc{
    [_timer release];
    [_mappers release];
    [_autoList release];
    [super dealloc];
}

@end
